from __future__ import annotations

import importlib
import logging
import sys
import types
import typing as t

from ...exceptions import MissingDependencyException

logger = logging.getLogger(__name__)


class LazyLoader(types.ModuleType):
    """
    LazyLoader module borrowed from Tensorflow
     https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/util/lazy_loader.py
     with a addition of "module caching". This will throw an
     exception if module cannot be imported.

    Lazily import a module, mainly to avoid pulling in large dependencies.
     `contrib`, and `ffmpeg` are examples of modules that are large and not always
     needed, and this allows them to only be loaded when they are used.
    """

    def __init__(
        self,
        local_name: str,
        parent_module_globals: dict[str, t.Any],
        name: str,
        warning: str | None = None,
        exc_msg: str | None = None,
        exc: type[Exception] = MissingDependencyException,
    ):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        self._warning = warning
        self._exc_msg = exc_msg
        self._exc = exc
        self._module: types.ModuleType | None = None

        super().__init__(str(name))

    def _load(self) -> types.ModuleType:
        """Load the module and insert it into the parent's globals."""
        # Import the target module and insert it into the parent's namespace
        try:
            module = importlib.import_module(self.__name__)
            self._parent_module_globals[self._local_name] = module
            # The additional add to sys.modules ensures library is actually loaded.
            sys.modules[self._local_name] = module
        except ModuleNotFoundError as err:
            raise self._exc(f"{self._exc_msg} (reason: {err})") from None

        # Emit a warning if one was specified
        if self._warning:
            logger.warning(self._warning)
            # Make sure to only warn once.
            self._warning = None

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item: t.Any) -> t.Any:
        if self._module is None:
            self._module = self._load()
        return getattr(self._module, item)

    def __dir__(self) -> t.List[str]:
        if self._module is None:
            self._module = self._load()
        return dir(self._module)



"""
User facing python APIs for building a OCI-complicant image.
"""

from __future__ import annotations

import logging
import os
import shutil
import sys
import typing as t

from simple_di import Provide
from simple_di import inject

from ._internal.configuration.containers import BentoMLContainer
from ._internal.container import build as _internal_build
from ._internal.container import (
    construct_containerfile as _internal_construct_containerfile,
)
from ._internal.container import get_backend
from ._internal.container import health
from ._internal.container import register_backend
from .exceptions import BentoMLException

if t.TYPE_CHECKING:
    from ._internal.bento import BentoStore
    from ._internal.container import DefaultBuilder
    from ._internal.container.base import ArgType
    from ._internal.tag import Tag
    from ._internal.types import PathType


logger = logging.getLogger(__name__)


@t.overload
def build(
    bento_tag: Tag | str,
    backend: t.Literal["docker"] = ...,
    image_tag: tuple[str, ...] | None = ...,
    features: t.Sequence[str] | None = ...,
    *,
    file: PathType | None = ...,
    context_path: PathType | None = ...,
    add_host: dict[str, str] | ArgType = ...,
    build_arg: dict[str, str] | ArgType = ...,
    cache_from: str | dict[str, str] | ArgType = ...,
    disable_content_trust: t.Literal[True, False] = ...,
    iidfile: PathType | None = ...,
    isolation: t.Literal["default", "process", "hyperv"] | None = ...,
    label: dict[str, str] | ArgType = ...,
    network: str | None = ...,
    no_cache: t.Literal[True, False] = ...,
    output: str | dict[str, str] | ArgType = ...,
    platform: str | ArgType = ...,
    progress: t.Literal["auto", "tty", "plain"] = ...,
    pull: t.Literal[True, False] = ...,
    quiet: t.Literal[True, False] = ...,
    secret: str | dict[str, str] | ArgType = ...,
    ssh: str | ArgType = ...,
    target: str | ArgType = ...,
    **kwargs: t.Any,
) -> t.Any:
    ...


@t.overload
def build(
    bento_tag: Tag | str,
    backend: t.Literal["buildctl"] = ...,
    image_tag: tuple[str] | None = ...,
    features: t.Sequence[str] | None = ...,
    *,
    file: PathType | None = ...,
    context_path: PathType | None = ...,
    output: str | dict[str, str] | ArgType = ...,
    progress: t.Literal["auto", "tty", "plain"] | ArgType = ...,
    trace: PathType | None = ...,
    local: dict[str, str] | ArgType = ...,
    frontend: str | None = ...,
    no_cache: t.Literal[True, False] = ...,
    export_cache: str | dict[str, str] | ArgType = ...,
    import_cache: str | dict[str, str] | ArgType = ...,
    secret: str | dict[str, str] | ArgType = ...,
    allow: str | ArgType = ...,
    ssh: str | ArgType = ...,
    metadata_file: PathType | None = ...,
    opt: tuple[str, ...] | dict[str, str | tuple[str, ...]] | None = ...,
    **kwargs: t.Any,
) -> t.Any:
    ...


@t.overload
def build(
    bento_tag: Tag | str,
    backend: t.Literal["buildx"] = ...,
    image_tag: tuple[str] | None = ...,
    features: t.Sequence[str] | None = ...,
    *,
    file: PathType | None = ...,
    tag: tuple[str] | None = ...,
    context_path: PathType = ...,
    add_host: dict[str, str] | ArgType = ...,
    allow: str | ArgType = ...,
    build_arg: dict[str, str] | ArgType = ...,
    build_context: dict[str, str] | ArgType = ...,
    builder: str | None = ...,
    cache_from: str | dict[str, str] | ArgType = ...,
    cache_to: str | dict[str, str] | ArgType = ...,
    cgroup_parent: PathType | None = ...,
    iidfile: PathType | None = ...,
    label: dict[str, str] | ArgType = ...,
    load: bool = True,
    metadata_file: PathType | None = ...,
    network: str | None = ...,
    no_cache: t.Literal[True, False] = ...,
    no_cache_filter: str | dict[str, str] | ArgType = ...,
    output: str | dict[str, str] | ArgType = ...,
    platform: str | ArgType = ...,
    progress: t.Literal["auto", "tty", "plain"] = "auto",
    pull: t.Literal[True, False] = ...,
    push: t.Literal[True, False] = ...,
    quiet: t.Literal[True, False] = ...,
    secret: str | dict[str, str] | ArgType = ...,
    shm_size: int | None = ...,
    ssh: str | ArgType = ...,
    target: str | None = ...,
    ulimit: str | dict[str, tuple[int, int]] | ArgType = ...,
    **kwargs: t.Any,
) -> t.Any:
    ...


@t.overload
def build(
    bento_tag: Tag | str,
    backend: t.Literal["nerdctl"] = ...,
    image_tag: tuple[str] | None = ...,
    features: t.Sequence[str] | None = ...,
    *,
    file: PathType | None = ...,
    tag: tuple[str] | None = ...,
    context_path: PathType = ".",
    build_arg: dict[str, str] | ArgType = ...,
    buildkit_host: str | None = ...,
    cache_from: str | dict[str, str] | ArgType = ...,
    cache_to: str | dict[str, str] | ArgType = ...,
    iidfile: PathType | None = ...,
    ipfs: t.Literal[True, False] = ...,
    label: dict[str, str] | ArgType = ...,
    no_cache: t.Literal[True, False] = ...,
    output: str | dict[str, str] | ArgType = ...,
    platform: str | ArgType = ...,
    progress: t.Literal["auto", "tty", "plain"] | ArgType = ...,
    quiet: t.Literal[True, False] = ...,
    rm: t.Literal[True, False] = ...,
    secret: str | dict[str, str] | ArgType = ...,
    ssh: str | ArgType = ...,
    target: str | None = ...,
    # global flags
    address: str | None = ...,
    host: str | None = ...,
    cgroup_manager: str | None = ...,
    cni_netconfpath: PathType | None = ...,
    cni_path: PathType | None = ...,
    data_root: PathType | None = ...,
    debug: t.Literal[True, False] = ...,
    debug_full: t.Literal[True, False] = ...,
    hosts_dir: str | ArgType = ...,
    insecure_registry: t.Literal[True, False] = ...,
    namespace: str | None = ...,
    snapshotter: str | None = ...,
    storage_driver: str | None = ...,
    **kwargs: t.Any,
) -> t.Any:
    ...


@t.overload
def build(
    bento_tag: Tag | str,
    backend: t.Literal["podman"] = ...,
    image_tag: tuple[str] | None = ...,
    features: t.Sequence[str] | None = ...,
    *,
    file: PathType | None = ...,
    context_path: PathType | None = ...,
    add_host: dict[str, str] | ArgType = ...,
    all_platforms: t.Literal[True, False] = ...,
    annotation: dict[str, str] | ArgType = ...,
    label: dict[str, str] | ArgType = ...,
    arch: str | None = ...,
    authfile: PathType | None = ...,
    build_arg: dict[str, str] | ArgType = ...,
    build_context: dict[str, str] | ArgType = ...,
    cache_from: str | None = ...,
    cache_to: str | None = ...,
    cache_ttl: str | None = ...,
    cap_add: str | ArgType = ...,
    cap_drop: str | ArgType = ...,
    cert_dir: PathType | None = ...,
    cgroup_parent: PathType | None = ...,
    cgroupns: str | None = ...,
    cpp_flag: ArgType = ...,
    cpu_period: int | None = ...,
    cpu_quota: int | None = ...,
    cpu_shares: int | None = ...,
    cpuset_cpus: str | None = ...,
    cpuset_mems: str | None = ...,
    creds: str | dict[str, str] | ArgType = ...,
    decryption_key: str | dict[str, str] | ArgType = ...,
    device: str | ArgType = ...,
    disable_compression: t.Literal[True, False] = ...,
    dns: str | None = ...,
    dns_option: str | ArgType = ...,
    dns_search: str | ArgType = ...,
    env: str | dict[str, str] | ArgType = ...,
    force_rm: t.Literal[True, False] = ...,
    format: str | t.Literal["docker", "oci"] | None = ...,
    hooks_dir: str | ArgType = ...,
    http_proxy: t.Literal[True, False] = ...,
    identity_label: t.Literal[True, False] = ...,
    ignorefile: PathType | None = ...,
    iidfile: PathType | None = ...,
    ipc: str | PathType | None = ...,
    isolation: str | None = ...,
    jobs: int | None = ...,
    layers: t.Literal[True, False] = ...,
    logfile: PathType | None = ...,
    manifest: str | None = ...,
    memory: str | None = ...,
    memory_swap: str | None = ...,
    network: str | None = ...,
    no_cache: t.Literal[True, False] = ...,
    no_hosts: t.Literal[True, False] = ...,
    omit_history: t.Literal[True, False] = ...,
    os: str | None = ...,
    os_feature: str | None = ...,
    os_version: str | None = ...,
    pid: PathType | None = ...,
    platform: str | ArgType = ...,
    output: str | dict[str, str] | ArgType = ...,
    pull: t.Literal[
        True, False, "always", "true", "missing", "never", "false", "newer"
    ] = False,
    quiet: t.Literal[True, False] = ...,
    rm: t.Literal[True, False] = ...,
    retry: int | None = ...,
    retry_delay: int | None = ...,
    runtime: PathType | None = ...,
    runtime_flag: str | dict[str, str] | ArgType = ...,
    secret: str | dict[str, str] | ArgType = ...,
    security_opt: str | ArgType = ...,
    shm_size: str | None = ...,
    sign_by: str | None = ...,
    skip_unused_stages: t.Literal[True, False] = ...,
    squash: t.Literal[True, False] = ...,
    squash_all: t.Literal[True, False] = ...,
    ssh: str | ArgType = ...,
    stdin: t.Literal[True, False] = ...,
    target: str | None = ...,
    timestamp: int | None = ...,
    tls_verify: t.Literal[True, False] = ...,
    ulimit: str | dict[str, tuple[int, int]] | ArgType = ...,
    unsetenv: str | ArgType = ...,
    userns: str | None = ...,
    userns_gid_map: str | tuple[str, str, str] | None = ...,
    userns_gid_map_group: str | None = ...,
    userns_uid_map: str | tuple[str, str, str] | None = ...,
    userns_uid_map_user: str | None = ...,
    uts: str | None = ...,
    variant: str | None = ...,
    volume: str | tuple[str, str, str] | None = ...,
    **kwargs: t.Any,
) -> t.Any:
    ...


@t.overload
def build(
    bento_tag: Tag | str,
    backend: t.Literal["buildah"] = ...,
    image_tag: tuple[str] | None = ...,
    features: t.Sequence[str] | None = ...,
    *,
    context_path: PathType = ...,
    file: PathType | None = ...,
    tag: tuple[str] | None = ...,
    add_host: dict[str, str] | ArgType = ...,
    annotation: dict[str, str] | ArgType = ...,
    label: dict[str, str] | ArgType = ...,
    arch: str | None = ...,
    authfile: PathType | None = ...,
    build_arg: dict[str, str] | ArgType = ...,
    cache_from: str | None = ...,
    cap_add: str | ArgType = ...,
    cap_drop: str | ArgType = ...,
    cert_dir: PathType | None = ...,
    cgroup_parent: PathType | None = ...,
    cni_config_dir: PathType | None = ...,
    cni_plugin_path: PathType | None = ...,
    compress: t.Literal[True, False] = ...,
    cpu_period: int | None = ...,
    cpu_quota: int | None = ...,
    cpu_shares: int | None = ...,
    cpuset_cpus: str | None = ...,
    cpuset_mems: str | None = ...,
    creds: str | dict[str, str] | ArgType = ...,
    decryption_key: str | dict[str, str] | ArgType = ...,
    device: str | ArgType = ...,
    disable_compression: t.Literal[True, False] = ...,
    dns: str | None = ...,
    dns_option: str | ArgType = ...,
    dns_search: str | ArgType = ...,
    force_rm: t.Literal[True, False] = ...,
    format: str | t.Literal["docker", "oci"] | None = ...,
    http_proxy: t.Literal[True, False] = ...,
    ignorefile: PathType | None = ...,
    iidfile: PathType | None = ...,
    ipc: str | PathType | None = ...,
    isolation: str | None = ...,
    jobs: int | None = ...,
    layers: t.Literal[True, False] = ...,
    logfile: PathType | None = ...,
    manifest: str | None = ...,
    memory: str | None = ...,
    memory_swap: str | None = ...,
    network: str | None = ...,
    no_cache: t.Literal[True, False] = ...,
    os: str | None = ...,
    pid: PathType | None = ...,
    platform: str | ArgType = ...,
    pull: t.Literal[True, False] = ...,
    pull_always: t.Literal[True, False] = ...,
    pull_never: t.Literal[True, False] = ...,
    quiet: t.Literal[True, False] = ...,
    rm: t.Literal[True, False] = ...,
    runtime: PathType | None = ...,
    runtime_flag: str | dict[str, str] | ArgType = ...,
    secret: str | dict[str, str] | ArgType = ...,
    security_opt: str | ArgType = ...,
    shm_size: str | None = ...,
    sign_by: str | None = ...,
    squash: t.Literal[True, False] = ...,
    ssh: str | ArgType = ...,
    stdin: t.Literal[True, False] = ...,
    target: str | None = ...,
    timestamp: int | None = ...,
    tls_verify: t.Literal[True, False] = ...,
    ulimit: str | dict[str, tuple[int, int]] | ArgType = ...,
    userns: str | None = ...,
    userns_gid_map_group: str | None = ...,
    userns_uid_map_user: str | None = ...,
    uts: str | None = ...,
    variant: str | None = ...,
    volume: str | tuple[str, str, str] | None = ...,
    **kwargs: t.Any,
) -> t.Any:
    ...


@inject
def build(
    bento_tag: Tag | str,
    backend: DefaultBuilder = "docker",
    image_tag: tuple[str] | None = None,
    features: t.Sequence[str] | None = None,
    _bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    **kwargs: t.Any,
) -> t.Any:
    """
    Build any given BentoML into a OCI-compliant image.

    .. code-block:: python

        import bentoml

        bento = bentoml.get("pytorch_vgg:latest")
        bentoml.container.build(bento, backend='podman', features=["grpc", "tracing"])

    Args:
        bento_tag: Bento tag in format of ``NAME:VERSION``
        backend: The backend to use for building the image. Current supported builder backends
                 include ``docker``, ``podman``, ``buildah``, ``nerdctl``, ``buildctl``, and ``buildx``.

                 .. note::

                     ``buildx`` is a syntatic sugar for ``docker buildx build``. See https://docs.docker.com/build/.
                     The reason for this is that ``buildx`` used to be the default behaviour of ``bentoml containerize``.
        image_tag: Optional additional image tag to apply to the built image.
        features: Optional features to include in the container file. See :ref:`concepts/bento:Python Packages`
                  for additional BentoML features.
        **kwargs: Additional keyword arguments to pass to the builder backend. Refer to the above overload
                  for each of the supported arguments per backend.
    """
    from ._internal.container import determine_container_tag

    # Run healthcheck
    if not health(backend):
        raise BentoMLException("Backend %s is not healthy." % backend)

    if "tag" not in kwargs:
        kwargs["tag"] = determine_container_tag(bento_tag, image_tag=image_tag)
    bento = _bento_store.get(bento_tag)

    logger.info("Building OCI-compliant image for %s with %s\n", bento.tag, backend)
    return _internal_build(bento_tag, backend, features=features, **kwargs)


@inject
def get_containerfile(
    bento_tag: Tag | str,
    output_path: str | None = None,
    enable_buildkit: bool = True,
    features: t.Sequence[str] | None = None,
    _bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
):
    """
    Returns the generated container file for a given Bento.

    Note that the container file (Dockerfile) inside the Bento is minimal, whereas
    this utility functions returns the container file that :ref:`bentoml containerize <reference/cli:containerize>` will
    be using.

    .. note::

        If ``output_path`` is not specified, then the contents of the container file
        will be printed out to ``sys.stderr``. If provided, then the final container file
        will be written to that given path.

    Args:
        bento_tag: Given tag for the bento.
        output_path: Optional output path to write the final container file to.
                     Note that if ``output_path`` is a directory, then the targeted file
                     will be ``output_path + os.sep + "<bento_tag>.dockerfile"``.
        enable_buildkit: Whether the container file contains BuildKit syntax.
        features: Optional features to include in the container file. See :ref:`concepts/bento:Python Packages`
                  for additional BentoML features.
    """
    bento = _bento_store.get(bento_tag)
    with _internal_construct_containerfile(
        bento,
        enable_buildkit=enable_buildkit,
        features=features,
        add_header=True,
    ) as (_, final_containerfile):
        if output_path is not None:
            if os.path.isdir(output_path):
                output_path = output_path = os.path.join(
                    output_path, f"{bento.tag.path().replace(os.sep,'_')}.dockerfile"
                )
            logger.info(
                "Writting Containerfile for '%s' to '%s'.", bento.tag, output_path
            )
            shutil.copyfile(final_containerfile, output_path)
            return
        # otherwise we will just write this to stderr.
        with open(final_containerfile, "r") as f:
            sys.stderr.write(f.read())


__all__ = ["build", "health", "register_backend", "get_backend", "get_containerfile"]



from __future__ import annotations

import typing as t
from abc import ABC
from abc import abstractmethod
from contextlib import contextmanager
from functools import wraps

from rich.console import Group
from rich.panel import Panel
from rich.progress import BarColumn
from rich.progress import DownloadColumn
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import TextColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn
from rich.progress import TransferSpeedColumn

from ..bento import Bento
from ..bento import BentoStore
from ..models import Model
from ..models import ModelStore
from ..tag import Tag

FILE_CHUNK_SIZE = 100 * 1024 * 1024  # 100Mb


class ObjectWrapper(object):
    def __getattr__(self, name: str) -> t.Any:
        return getattr(self._wrapped, name)

    def __setattr__(self, name: str, value: t.Any) -> None:
        return setattr(self._wrapped, name, value)

    def wrapper_getattr(self, name: str):
        """Actual `self.getattr` rather than self._wrapped.getattr"""
        return getattr(self, name)

    def wrapper_setattr(self, name: str, value: t.Any) -> None:
        """Actual `self.setattr` rather than self._wrapped.setattr"""
        return object.__setattr__(self, name, value)

    def __init__(self, wrapped: t.Any):
        """
        Thin wrapper around a given object
        """
        self.wrapper_setattr("_wrapped", wrapped)


class CallbackIOWrapper(ObjectWrapper):
    def __init__(
        self,
        callback: t.Callable[[int], None],
        stream: t.BinaryIO,
        method: t.Literal["read", "write"] = "read",
    ):
        """
        Wrap a given `file`-like object's `read()` or `write()` to report
        lengths to the given `callback`
        """
        super().__init__(stream)
        func = getattr(stream, method)
        if method == "write":

            @wraps(func)
            def write(data: t.Union[bytes, bytearray], *args: t.Any, **kwargs: t.Any):
                res = func(data, *args, **kwargs)
                callback(len(data))
                return res

            self.wrapper_setattr("write", write)
        elif method == "read":

            @wraps(func)
            def read(*args: t.Any, **kwargs: t.Any):
                data = func(*args, **kwargs)
                callback(len(data))
                return data

            self.wrapper_setattr("read", read)
        else:
            raise KeyError("Can only wrap read/write methods")


class CloudClient(ABC):
    log_progress = Progress(TextColumn("{task.description}"))

    spinner_progress = Progress(
        TextColumn("  "),
        TimeElapsedColumn(),
        TextColumn("[bold purple]{task.fields[action]}"),
        SpinnerColumn("simpleDots"),
    )

    transmission_progress = Progress(
        TextColumn("[bold blue]{task.description}", justify="right"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    )

    progress_group = Group(
        Panel(Group(log_progress, spinner_progress)), transmission_progress
    )

    @contextmanager
    def spin(self, *, text: str):
        task_id = self.spinner_progress.add_task("", action=text)
        try:
            yield
        finally:
            self.spinner_progress.stop_task(task_id)
            self.spinner_progress.update(task_id, visible=False)

    @abstractmethod
    def push_model(
        self,
        model: Model,
        *,
        force: bool = False,
        threads: int = 10,
        context: str | None = None,
    ):
        pass

    @abstractmethod
    def push_bento(
        self,
        bento: Bento,
        *,
        force: bool = False,
        threads: int = 10,
        context: str | None = None,
    ):
        pass

    @abstractmethod
    def pull_model(
        self,
        tag: str | Tag,
        *,
        force: bool = False,
        context: str | None = None,
        model_store: ModelStore,
    ) -> Model:
        pass

    @abstractmethod
    def pull_bento(
        self,
        tag: str | Tag,
        *,
        force: bool = False,
        context: str | None = None,
        bento_store: BentoStore,
    ) -> Bento:
        pass



from __future__ import annotations

import json
import typing as t
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING

import attr
import cattr
from dateutil.parser import parse

time_format = "%Y-%m-%d %H:%M:%S.%f"


def datetime_encoder(time_obj: t.Optional[datetime]) -> t.Optional[str]:
    if not time_obj:
        return None
    return time_obj.strftime(time_format)


def datetime_decoder(datetime_str: t.Optional[str], _: t.Any) -> t.Optional[datetime]:
    if not datetime_str:
        return None
    return parse(datetime_str)


def dict_options_converter(
    options_type: type[t.Any],
) -> t.Callable[[type[t.Any] | dict[str, t.Any]], t.Any]:
    def _converter(value: type[t.Any] | dict[str, t.Any] | None) -> options_type:
        if value is None:
            return options_type()
        if isinstance(value, dict):
            return options_type(**value)
        return value

    return _converter


cloud_converter = cattr.Converter()

cloud_converter.register_unstructure_hook(datetime, datetime_encoder)
cloud_converter.register_structure_hook(datetime, datetime_decoder)


T = t.TypeVar("T")


def schema_from_json(json_content: str, cls: t.Type[T]) -> T:
    dct = json.loads(json_content)
    return cloud_converter.structure(dct, cls)


def schema_to_json(obj: t.Any) -> str:
    res = cloud_converter.unstructure(obj, obj.__class__)
    return json.dumps(res)


def schema_from_object(obj: t.Any, cls: t.Type[T]) -> T:
    return cloud_converter.structure(obj, cls)


@attr.define
class BaseSchema:
    uid: str
    created_at: datetime
    updated_at: t.Optional[datetime]
    deleted_at: t.Optional[datetime]


@attr.define
class BaseListSchema:
    start: int
    count: int
    total: int


class ResourceType(Enum):
    USER = "user"
    ORG = "organization"
    CLUSTER = "cluster"
    HostCluster = "host_cluster"
    BENTO_REPOSITORY = "bento_repository"
    BENTO = "bento"
    MODEL_REPOSITORY = "model_repository"
    MODEL = "model"
    DEPLOYMENT = "deployment"
    DEPLOYMENT_REVISION = "deployment_revision"
    TERMINAL_RECORD = "terminal_record"
    LABEL = "label"
    API_TOKEN = "api_token"
    YATAI_COMPONENT = "yatai_component"
    LimitGroup = "limit_group"
    ResourceInstance = "resource_instance"


@attr.define
class ResourceSchema(BaseSchema):
    name: str
    resource_type: ResourceType
    labels: t.List[LabelItemSchema]


@attr.define
class UserSchema:
    name: str
    email: str
    first_name: str
    last_name: str

    def get_name(self) -> str:
        if not self.first_name and not self.last_name:
            return self.name
        return f"{self.first_name} {self.last_name}".strip()


@attr.define
class OrganizationSchema(ResourceSchema):
    description: str


@attr.define
class OrganizationListSchema(BaseListSchema):
    items: t.List[OrganizationSchema]


@attr.define
class ClusterSchema(ResourceSchema):
    description: str
    creator: UserSchema


@attr.define
class ClusterConfigSchema:
    default_deployment_kube_namespace: str
    ingress_ip: str
    aws: str


@attr.define
class ClusterListSchema(BaseListSchema):
    items: t.List[ClusterSchema]


@attr.define
class CreateBentoRepositorySchema:
    name: str
    description: str


class BentoImageBuildStatus(Enum):
    PENDING = "pending"
    BUILDING = "building"
    SUCCESS = "success"
    FAILED = "failed"


class BentoUploadStatus(Enum):
    PENDING = "pending"
    BUILDING = "uploading"
    SUCCESS = "success"
    FAILED = "failed"


@attr.define
class BentoApiSchema:
    route: str
    doc: str
    input: str
    output: str


@attr.define
class BentoRunnerResourceSchema:
    cpu: t.Optional[t.Any]
    nvidia_gpu: t.Optional[t.Any]
    custom_resources: t.Optional[t.Any]


@attr.define
class BentoRunnerSchema:
    name: str
    runnable_type: t.Optional[str]
    models: t.Optional[t.List[str]]
    resource_config: t.Optional[BentoRunnerResourceSchema]


@attr.define
class BentoManifestSchema:
    service: str
    bentoml_version: str
    size_bytes: int
    apis: t.Dict[str, BentoApiSchema] = attr.field(factory=dict)
    models: t.List[str] = attr.field(factory=list)
    runners: t.Optional[t.List[BentoRunnerSchema]] = attr.field(factory=list)


if TYPE_CHECKING:
    TransmissionStrategy = t.Literal["presigned_url", "proxy"]
else:
    TransmissionStrategy = str


@attr.define
class BentoSchema(ResourceSchema):
    description: str
    version: str
    image_build_status: BentoImageBuildStatus
    upload_status: BentoUploadStatus
    upload_finished_reason: str
    presigned_upload_url: str
    presigned_download_url: str
    manifest: BentoManifestSchema
    transmission_strategy: t.Optional[TransmissionStrategy] = attr.field(default=None)
    upload_id: t.Optional[str] = attr.field(default=None)

    upload_started_at: t.Optional[datetime] = attr.field(default=None)
    upload_finished_at: t.Optional[datetime] = attr.field(default=None)
    build_at: datetime = attr.field(factory=datetime.now)


@attr.define
class BentoRepositorySchema(ResourceSchema):
    description: str
    latest_bento: t.Optional[BentoSchema]


@attr.define
class BentoWithRepositorySchema(BentoSchema):
    repository: BentoRepositorySchema = attr.field(default=None)


@attr.define
class BentoWithRepositoryListSchema(BaseListSchema):
    items: t.List[BentoWithRepositorySchema] = attr.field(factory=list)


@attr.define
class CreateBentoSchema:
    description: str
    version: str
    manifest: BentoManifestSchema
    build_at: datetime = attr.field(factory=datetime.now)
    labels: t.List[LabelItemSchema] = attr.field(factory=list)


@attr.define
class UpdateBentoSchema:
    manifest: t.Optional[BentoManifestSchema] = attr.field(default=None)
    labels: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)


@attr.define
class BentoFullSchema(BentoWithRepositorySchema):
    models: t.List[ModelWithRepositorySchema] = attr.field(factory=list)


@attr.define
class PreSignMultipartUploadUrlSchema:
    upload_id: str
    part_number: int


@attr.define
class CompletePartSchema:
    part_number: int
    etag: str


@attr.define
class CompleteMultipartUploadSchema:
    parts: t.List[CompletePartSchema]
    upload_id: str


@attr.define
class FinishUploadBentoSchema:
    status: t.Optional[BentoUploadStatus]
    reason: t.Optional[str]


@attr.define
class CreateModelRepositorySchema:
    name: str
    description: str


class ModelImageBuildStatus(Enum):
    PENDING = "pending"
    BUILDING = "building"
    SUCCESS = "success"
    FAILED = "failed"


class ModelUploadStatus(Enum):
    PENDING = "pending"
    BUILDING = "uploading"
    SUCCESS = "success"
    FAILED = "failed"


@attr.define
class ModelManifestSchema:
    module: str
    api_version: str
    bentoml_version: str
    size_bytes: int
    metadata: t.Dict[str, t.Any] = attr.field(factory=dict)
    context: t.Dict[str, t.Any] = attr.field(factory=dict)
    options: t.Dict[str, t.Any] = attr.field(factory=dict)


@attr.define
class ModelSchema(ResourceSchema):
    description: str
    version: str
    image_build_status: ModelImageBuildStatus
    upload_status: ModelUploadStatus
    upload_finished_reason: str
    presigned_upload_url: str
    presigned_download_url: str
    manifest: ModelManifestSchema

    transmission_strategy: t.Optional[TransmissionStrategy] = attr.field(default=None)
    upload_id: t.Optional[str] = attr.field(default=None)

    upload_started_at: t.Optional[datetime] = attr.field(default=None)
    upload_finished_at: t.Optional[datetime] = attr.field(default=None)
    build_at: datetime = attr.field(factory=datetime.now)


@attr.define
class ModelRepositorySchema(ResourceSchema):
    description: str
    latest_model: t.Optional[ModelSchema]


@attr.define
class ModelWithRepositorySchema(ModelSchema):
    repository: ModelRepositorySchema = attr.field(default=None)


@attr.define
class ModelWithRepositoryListSchema(BaseListSchema):
    items: t.List[ModelWithRepositorySchema] = attr.field(factory=list)


@attr.define
class CreateModelSchema:
    description: str
    version: str
    manifest: ModelManifestSchema
    build_at: datetime = attr.field(factory=datetime.now)
    labels: t.List[LabelItemSchema] = attr.field(factory=list)


@attr.define
class FinishUploadModelSchema:
    status: t.Optional[ModelUploadStatus]
    reason: t.Optional[str]


@attr.define
class BentoRepositoryListSchema(BaseListSchema):
    items: t.List[BentoRepositorySchema]


@attr.define
class BentoListSchema(BaseListSchema):
    items: t.List[BentoSchema]


class DeploymentTargetCanaryRuleType(Enum):
    WEIGHT = "weight"
    HEADER = "header"
    COOKIE = "cookie"


@attr.define
class DeploymentTargetCanaryRule:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    type: DeploymentTargetCanaryRuleType
    weight: int
    header: str
    cookie: str
    header_value: str


@attr.define
class ApiServerBentoDeploymentOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    monitorExporter: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)


@attr.define
class RunnerBentoDeploymentOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    extraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)


@attr.define
class BentoRequestOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    imageBuildTimeout: int = attr.field(default=None)
    imageBuilderExtraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(
        default=None
    )
    imageBuilderExtraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    imageBuilderExtraContainerEnv: t.Optional[t.List[t.Dict[str, t.Any]]] = attr.field(
        default=None
    )
    imageBuilderContainerResources: t.Optional[t.Dict[str, t.Any]] = attr.field(
        default=None
    )
    dockerConfigJsonSecretName: t.Optional[str] = attr.field(default=None)
    downloaderContainerEnvFrom: t.Optional[t.Dict[str, t.Any]] = attr.field(
        default=None
    )


@attr.define
class LabelItemSchema:
    key: str
    value: str


class HPAMetricType(Enum):
    MEMORY = "memory"
    CPU = "cpu"
    GPU = "gpu"
    QPS = "qps"


@attr.define
class HPAMetric:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    type: HPAMetricType  # enum
    value: t.Any  # resource.Quantity


class HPAScaleBehavior(Enum):
    DISABLED = "disabled"
    STABLE = "stable"
    FAST = "fast"


@attr.define
class HPAPolicy:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    metrics: t.Optional[t.List[HPAMetric]] = attr.field(default=None)
    scale_down_behavior: t.Optional[HPAScaleBehavior] = attr.field(default=None)
    scale_up_behavior: t.Optional[HPAScaleBehavior] = attr.field(default=None)


@attr.define
class DeploymentTargetHPAConf:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    cpu: t.Optional[int] = attr.field(default=None)
    gpu: t.Optional[int] = attr.field(default=None)
    memory: t.Optional[str] = attr.field(default=None)
    qps: t.Optional[int] = attr.field(default=None)
    min_replicas: t.Optional[int] = attr.field(default=None)
    max_replicas: t.Optional[int] = attr.field(default=None)
    policy: t.Optional[HPAPolicy] = attr.field(default=None)


@attr.define
class DeploymentTargetResourceItem:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    cpu: t.Optional[str] = attr.field(default=None)
    memory: t.Optional[str] = attr.field(default=None)
    gpu: t.Optional[str] = attr.field(default=None)
    custom: t.Optional[t.Dict[str, str]] = attr.field(default=None)


@attr.define
class DeploymentTargetResources:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    requests: t.Optional[DeploymentTargetResourceItem] = attr.field(default=None)
    limits: t.Optional[DeploymentTargetResourceItem] = attr.field(default=None)


@attr.define
class RequestQueueConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    enabled: t.Optional[bool] = attr.field(default=None)
    max_consume_concurrency: t.Optional[int] = attr.field(default=None)


@attr.define
class TrafficControlConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    timeout: t.Optional[str] = attr.field(default=None)
    request_queue: t.Optional[RequestQueueConfig] = attr.field(default=None)


class DeploymentStrategy(Enum):
    RollingUpdate = "RollingUpdate"
    Recreate = "Recreate"
    RampedSlowRollout = "RampedSlowRollout"
    BestEffortControlledRollout = "BestEffortControlledRollout"


@attr.define
class DeploymentTargetRunnerConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    resource_instance: t.Optional[str] = attr.field(default=None)
    resources: t.Optional[DeploymentTargetResources] = attr.field(default=None)
    hpa_conf: t.Optional[DeploymentTargetHPAConf] = attr.field(default=None)
    envs: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)
    enable_stealing_traffic_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_pod_receive_production_traffic: t.Optional[bool] = attr.field(
        default=None
    )
    deployment_strategy: t.Optional[DeploymentStrategy] = attr.field(default=None)
    bento_deployment_overrides: t.Optional[RunnerBentoDeploymentOverrides] = attr.field(
        default=None
    )
    traffic_control: t.Optional[TrafficControlConfig] = attr.field(default=None)
    deployment_cold_start_wait_timeout: t.Optional[int] = attr.field(default=None)


class DeploymentTargetType(Enum):
    STABLE = "stable"
    CANARY = "canary"


@attr.define
class DeploymentTargetConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    resources: DeploymentTargetResources = attr.field(
        default=None, converter=dict_options_converter(DeploymentTargetResources)
    )
    kubeResourceUid: str = attr.field(default="")  # empty str
    kubeResourceVersion: str = attr.field(default="")
    resource_instance: t.Optional[str] = attr.field(default=None)
    hpa_conf: t.Optional[DeploymentTargetHPAConf] = attr.field(default=None)
    envs: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)
    runners: t.Optional[t.Dict[str, DeploymentTargetRunnerConfig]] = attr.field(
        default=None
    )
    access_control: t.Optional[str] = attr.field(default=None)
    enable_ingress: t.Optional[bool] = attr.field(default=None)  # false for enables
    enable_stealing_traffic_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_pod_receive_production_traffic: t.Optional[bool] = attr.field(
        default=None
    )
    deployment_strategy: t.Optional[DeploymentStrategy] = attr.field(
        default=None
    )  # Specific
    bento_deployment_overrides: t.Optional[
        ApiServerBentoDeploymentOverrides
    ] = attr.field(default=None)
    bento_request_overrides: t.Optional[BentoRequestOverrides] = attr.field(
        default=None
    )  # Put into image builder
    traffic_control: t.Optional[TrafficControlConfig] = attr.field(default=None)
    deployment_cold_start_wait_timeout: t.Optional[int] = attr.field(default=None)


@attr.define
class CreateDeploymentTargetSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    type: DeploymentTargetType  # stable by default
    bento_repository: str
    bento: str
    config: DeploymentTargetConfig
    canary_rules: t.Optional[t.List[DeploymentTargetCanaryRule]] = attr.field(
        default=None
    )


class DeploymentStatus(Enum):
    Unknown = "unknown"
    NonDeployed = "non-deployed"
    Running = "running"
    Unhealthy = "unhealthy"
    Failed = "failed"
    Deploying = "deploying"
    Terminating = "terminating"
    Terminated = "terminated"
    ImageBuilding = "image-building"
    ImageBuildFailed = "image-build-failed"
    ImageBuildSucceeded = "image-build-succeeded"


@attr.define
class DeploymentSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    creator: UserSchema
    cluster: ClusterSchema
    status: DeploymentStatus
    kube_namespace: str
    latest_revision: t.Optional[DeploymentRevisionSchema] = attr.field(
        default=None
    )  # Delete returns no latest revision
    mode: t.Optional[DeploymentMode] = attr.field(default=None)


@attr.define
class DeploymentTargetSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    creator: UserSchema
    type: DeploymentTargetType
    bento: BentoFullSchema
    config: DeploymentTargetConfig
    canary_rules: t.Optional[t.List[DeploymentTargetCanaryRule]] = attr.field(
        default=None
    )


class DeploymentRevisionStatus(Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"


@attr.define
class DeploymentRevisionSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    creator: UserSchema
    status: DeploymentRevisionStatus
    targets: t.List[DeploymentTargetSchema]


@attr.define
class ResourceInstanceConfigSchema:
    group: str
    resources: DeploymentTargetResources
    price: str
    node_selectors: t.Optional[t.Dict[str, str]] = attr.field(factory=dict)


@attr.define
class ResourceInstanceSchema(ResourceSchema):
    display_name: str
    description: str
    config: ResourceInstanceConfigSchema


@attr.define
class ClusterFullSchema(ClusterSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    organization: OrganizationSchema
    kube_config: str
    config: ClusterConfigSchema
    grafana_root_path: str
    resource_instances: t.List[ResourceInstanceSchema]


@attr.define
class DeploymentListSchema(BaseListSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    items: t.List[DeploymentSchema]


class DeploymentMode(Enum):
    Deployment = "deployment"
    Function = "function"


@attr.define
class UpdateDeploymentSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    targets: t.List[CreateDeploymentTargetSchema]
    mode: t.Optional[DeploymentMode] = attr.field(default=None)
    labels: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)
    description: t.Optional[str] = attr.field(default=None)
    do_not_deploy: t.Optional[bool] = attr.field(default=None)


@attr.define
class CreateDeploymentSchema(UpdateDeploymentSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    name: str = attr.field(default=None)
    kube_namespace: t.Optional[str] = attr.field(default=None)


@attr.define
class FullDeploymentSchema(CreateDeploymentSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = True
    cluster_name: t.Optional[str] = attr.field(default=None)



from __future__ import annotations

import logging
import os
import re
import typing as t
import urllib.parse
from abc import ABC
from abc import abstractmethod

import fs
import fs.copy
import fs.errors
import fs.mirror
import fs.opener
import fs.opener.errors
import fs.tempfs
from fs import open_fs
from fs.base import FS

T = t.TypeVar("T", bound="Exportable")


uriSchemeRe = re.compile(r".*[^\\](?=://)")


class Exportable(ABC):
    @staticmethod
    @abstractmethod
    def _export_ext() -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _export_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _fs(self) -> FS:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_fs(cls: t.Type[T], item_fs: FS) -> T:
        raise NotImplementedError

    @classmethod
    def guess_format(cls, path: str) -> str:
        _, ext = fs.path.splitext(path)

        if ext == "":
            return cls._export_ext()

        ext = ext[1:]
        if ext in [cls._export_ext(), "gz", "xz", "bz2", "tar", "zip"]:
            return ext
        else:
            return cls._export_ext()

    @classmethod
    def import_from(
        cls: t.Type[T],
        path: str,
        input_format: t.Optional[str] = None,
        *,
        protocol: t.Optional[str] = None,
        user: t.Optional[str] = None,
        passwd: t.Optional[str] = None,
        params: t.Optional[t.Dict[str, str]] = None,
        subpath: t.Optional[str] = None,
    ) -> T:
        try:
            parsedurl = fs.opener.parse(path)
        except fs.opener.errors.ParseError:
            if protocol is None:
                protocol = "osfs"
                resource: str = path if os.sep == "/" else path.replace(os.sep, "/")
            else:
                resource: str = ""
        else:
            if any(v is not None for v in [protocol, user, passwd, params, subpath]):
                raise ValueError(
                    "An FS URL was passed as the output path; all additional information should be passed as part of the URL."
                )

            protocol = parsedurl.protocol  # type: ignore (FS types)
            user = parsedurl.username  # type: ignore (FS types)
            passwd = parsedurl.password  # type: ignore (FS types)
            resource: str = parsedurl.resource  # type: ignore (FS types)
            params = parsedurl.params  # type: ignore (FS types)
            subpath = parsedurl.path  # type: ignore (FS types)

        if protocol not in fs.opener.registry.protocols:  # type: ignore (FS types)
            if protocol == "s3":
                raise ValueError(
                    "Tried to open an S3 url but the protocol is not registered; did you 'pip install \"bentoml[aws]\"'?"
                )
            else:
                raise ValueError(
                    f"Unknown or unsupported protocol {protocol}. Some supported protocols are 'ftp', 's3', and 'osfs'."
                )

        isOSPath = protocol in [
            "temp",
            "osfs",
            "userdata",
            "userconf",
            "sitedata",
            "siteconf",
            "usercache",
            "userlog",
        ]

        isCompressedPath = protocol in ["tar", "zip"]

        if input_format is not None:
            if isCompressedPath:
                raise ValueError(
                    f"A {protocol} path was passed along with an input format ({input_format}); only pass one or the other."
                )
        else:
            # try to guess output format if it hasn't been passed
            input_format = cls.guess_format(resource if subpath is None else subpath)

        if subpath is None:
            if isCompressedPath:
                subpath = ""
            else:
                resource, subpath = fs.path.split(resource)

        if isOSPath:
            # we can safely ignore user / passwd / params
            # get the path on the system so we can skip the copy step
            try:
                rsc_dir = fs.open_fs(f"{protocol}://{resource}")
                # if subpath is root, we can just open directly
                if (subpath == "" or subpath == "/") and input_format == "folder":
                    return cls.from_fs(rsc_dir)

                # if subpath is specified, we need to create our own temp fs and mirror that subpath
                path = rsc_dir.getsyspath(subpath)
                res = cls._from_compressed(path, input_format)
            except fs.errors.CreateFailed:
                raise ValueError("Directory does not exist")
        else:
            userblock = ""
            if user is not None or passwd is not None:
                if user is not None:
                    userblock += urllib.parse.quote(user)
                if passwd is not None:
                    userblock += ":" + urllib.parse.quote(passwd)
                userblock += "@"

            resource = urllib.parse.quote(resource)

            query = ""
            if params is not None:
                query = "?" + urllib.parse.urlencode(params)

            input_uri = f"{protocol}://{userblock}{resource}{query}"

            if isCompressedPath:
                input_fs = fs.open_fs(input_uri)
                res = cls.from_fs(input_fs)
            else:
                filename = fs.path.basename(subpath)
                tempfs = fs.tempfs.TempFS(identifier=f"bentoml-{filename}-import")
                if input_format == "folder":
                    input_fs = fs.open_fs(input_uri)
                    fs.copy.copy_dir(input_fs, subpath, tempfs, filename)
                else:
                    input_fs = fs.open_fs(input_uri)
                    fs.copy.copy_file(input_fs, subpath, tempfs, filename)

                res = cls._from_compressed(tempfs.getsyspath(filename), input_format)

            input_fs.close()

        return res

    def export(
        self,
        path: str,
        output_format: t.Optional[str] = None,
        *,
        protocol: t.Optional[str] = None,
        user: t.Optional[str] = None,
        passwd: t.Optional[str] = None,
        params: t.Optional[t.Dict[str, str]] = None,
        subpath: t.Optional[str] = None,
    ) -> str:
        try:
            parsedurl = fs.opener.parse(path)
        except fs.opener.errors.ParseError:  # type: ignore (incomplete FS types)
            if protocol is None:
                protocol = "osfs"
                resource = path if os.sep == "/" else path.replace(os.sep, "/")
            else:
                resource = path
        else:
            if any(v is not None for v in [protocol, user, passwd, params, subpath]):
                raise ValueError(
                    "An FS URL was passed as the output path; all additional information should be passed as part of the URL."
                )

            protocol = parsedurl.protocol  # type: ignore (incomplete FS types)
            user = parsedurl.username  # type: ignore (incomplete FS types)
            passwd = parsedurl.password  # type: ignore (incomplete FS types)
            resource: str = parsedurl.resource  # type: ignore (incomplete FS types)
            params = parsedurl.params  # type: ignore (incomplete FS types)
            subpath = parsedurl.path  # type: ignore (incomplete FS types)

        if protocol not in fs.opener.registry.protocols:  # type: ignore (incomplete FS types)
            if protocol == "s3":
                raise ValueError(
                    "Tried to open an S3 url but the protocol is not registered; did you install fs-s3fs?"
                )
            else:
                raise ValueError(
                    f"Unknown or unsupported protocol {protocol}. Some supported protocols are 'ftp', 's3', and 'osfs'."
                )

        isOSPath = protocol in [
            "temp",
            "osfs",
            "userdata",
            "userconf",
            "sitedata",
            "siteconf",
            "usercache",
            "userlog",
        ]

        isCompressedPath = protocol in ["tar", "zip"]

        if output_format is not None:
            if isCompressedPath:
                raise ValueError(
                    f"A {protocol} path was passed along with an output format ({output_format}); only pass one or the other."
                )
        else:
            # try to guess output format if it hasn't been passed
            output_format = self.guess_format(resource if subpath is None else subpath)

        # use combine here instead of join because we don't want to fold subpath into path.
        fullpath = resource if subpath is None else fs.path.combine(resource, subpath)

        if fullpath == "" or (not output_format == "folder" and fullpath[-1] == "/"):
            subpath = "" if subpath is None else subpath
            # if the path has no filename or is empty, we generate a filename from the bento tag
            subpath = fs.path.join(subpath, self._export_name)
            if output_format in ["gz", "xz", "bz2", "tar", "zip"]:
                subpath += "." + output_format
        elif isOSPath:
            try:
                dirfs = fs.open_fs(path)
            except fs.errors.CreateFailed:
                pass
            else:
                if subpath is None or dirfs.isdir(subpath):
                    subpath = "" if subpath is None else subpath
                    # path exists and is a directory, we will create a file inside that directory
                    subpath = fs.path.join(subpath, self._export_name)
                    if output_format in ["gz", "xz", "bz2", "tar", "zip"]:
                        subpath += "." + output_format

        if subpath is None:
            if isCompressedPath:
                subpath = ""
            else:
                resource, subpath = fs.path.split(resource)

        # as a special case, if the output format is the default we will always append the relevant
        # extension if the output filename does not have it
        if (
            output_format == self._export_ext()
            and fs.path.splitext(subpath)[1] != "." + self._export_ext()
        ):
            logging.info(
                f"adding {self._export_ext} because ext is {fs.path.splitext(subpath)[1]}"
            )
            subpath += "." + self._export_ext()

        if isOSPath:
            # we can safely ignore user / passwd / params
            # get the path on the system so we can skip the copy step
            try:
                rsc_dir = fs.open_fs(f"{protocol}://{resource}")
                path = rsc_dir.getsyspath(subpath)
                self._compress(path, output_format)
            except fs.errors.CreateFailed:
                raise ValueError(
                    f"Output directory '{protocol}://{resource}' does not exist."
                )
        else:
            userblock = ""
            if user is not None or passwd is not None:
                if user is not None:
                    userblock += urllib.parse.quote(user)
                if passwd is not None:
                    userblock += ":" + urllib.parse.quote(passwd)
                userblock += "@"

            resource = urllib.parse.quote(resource)

            query = ""
            if params is not None:
                query = "?" + urllib.parse.urlencode(params)

            dest_uri = f"{protocol}://{userblock}{resource}{query}"

            if isCompressedPath:
                destfs = fs.open_fs(dest_uri, writeable=True, create=True)
                fs.mirror.mirror(self._fs, destfs, copy_if_newer=False)
            else:
                tempfs = fs.tempfs.TempFS(
                    identifier=f"bentoml-{self._export_name}-export"
                )
                filename = fs.path.basename(subpath)
                self._compress(tempfs.getsyspath(filename), output_format)

                if output_format == "folder":
                    destfs = fs.open_fs(dest_uri)
                    fs.copy.copy_dir(tempfs, filename, destfs, subpath)
                else:
                    destfs = fs.open_fs(dest_uri)
                    fs.copy.copy_file(tempfs, filename, destfs, subpath)

                tempfs.close()

            destfs.close()

        return path

    def _compress(self, path: str, output_format: str):
        if output_format in ["gz", "xz", "bz2", "tar", self._export_ext()]:
            from fs.tarfs import WriteTarFS

            compression = output_format
            if compression == "tar":
                compression = None
            if compression == self._export_ext():
                compression = "xz"
            out_fs = WriteTarFS(path, compression)
        elif output_format == "zip":
            from fs.zipfs import WriteZipFS

            out_fs = WriteZipFS(path)
        elif output_format == "folder":
            out_fs: FS = open_fs(path, writeable=True, create=True)
        else:
            raise ValueError(f"Unsupported format {output_format}")

        fs.mirror.mirror(self._fs, out_fs, copy_if_newer=False)
        out_fs.close()

    @classmethod
    def _from_compressed(cls: t.Type[T], path: str, input_format: str) -> T:
        if input_format in ["gz", "xz", "bz2", "tar", cls._export_ext()]:
            from fs.tarfs import ReadTarFS

            compression = input_format
            if compression == "tar":
                compression = None
            if compression == cls._export_ext():
                compression = "xz"
            ret_fs = ReadTarFS(path)
        elif input_format == "zip":
            from fs.zipfs import ReadZipFS

            ret_fs = ReadZipFS(path)
        elif input_format == "folder":
            ret_fs: FS = open_fs(path)
        else:
            raise ValueError(f"Unsupported format {input_format}")

        return cls.from_fs(ret_fs)



from __future__ import annotations

import contextlib
import logging
import logging.config
import typing as t

from simple_di import Provide
from simple_di import inject

from ..configuration.containers import BentoMLContainer
from ..types import LazyType
from .base import MT
from .base import MonitorBase
from .base import NoOpMonitor
from .default import DefaultMonitor

logger = logging.getLogger(__name__)

_MONITOR_INSTANCES: dict[str, MonitorBase[t.Any]] = {}  # cache of monitors


@t.overload
@contextlib.contextmanager
def monitor(
    name: str | t.Any,
    monitor_class: DefaultMonitor = ...,
    monitor_options: dict[str, t.Any] | None = ...,
) -> t.Generator[DefaultMonitor, None, None]:
    ...


@t.overload
@contextlib.contextmanager
def monitor(
    name: str | t.Any,
    monitor_class: str = ...,
    monitor_options: dict[str, t.Any] | None = ...,
) -> t.Generator[MonitorBase[t.Any], None, None]:
    ...


@t.overload
@contextlib.contextmanager
def monitor(
    name: str | t.Any,
    monitor_class: None = ...,
    monitor_options: dict[str, t.Any] | None = ...,
) -> t.Generator[MonitorBase[t.Any], None, None]:
    ...


@contextlib.contextmanager
@inject
def monitor(
    name: str,
    monitor_class: type[MT]
    | str
    | None = Provide[BentoMLContainer.config.monitoring.type],
    monitor_options: dict[str, t.Any]
    | None = Provide[BentoMLContainer.config.monitoring.options],
) -> t.Generator[MT | MonitorBase[t.Any], None, None]:
    """
    Context manager for monitoring.

    :param name: name of the monitor
    :param monitor_class: class of the monitor, can be a string or a class
        example:
        - default
        - otlp
        - "bentoml.monitoring.prometheus.PrometheusMonitor"
    :param monitor_options: options for the monitor

    :return: a monitor instance

    Example::

        with bentoml.monitor("my_monitor") as m:
            m.log(1, "x", "feature", "numerical")
            m.log(2, "y", "feature", "numerical")
            m.log(3, "z", "feature", "numerical")
            m.log(4, "prediction", "prediction", "numerical")

        # or
        with bentoml.monitor("my_monitor") as m:
            m.log_batch([1, 2, 3], "x", "feature", "numerical")
            m.log_batch([4, 5, 6], "y", "feature", "numerical")
            m.log_batch([7, 8, 9], "z", "feature", "numerical")
            m.log_batch([10, 11, 12], "prediction", "prediction", "numerical")

        this will log the following data:
        {
            "timestamp": "2021-09-01T00:00:00",
            "request_id": "1234567890",
            "x": 1,
            "y": 2,
            "z": 3,
            "prediction": 4,
        }
        and the following schema:
        [
            {"name": "timestamp", "role": "time", "type": "datetime"},
            {"name": "request_id", "role": "request_id", "type": "string"},
            {"name": "x", "role": "feature", "type": "numerical"},
            {"name": "y", "role": "feature", "type": "numerical"},
            {"name": "z", "role": "feature", "type": "numerical"},
            {"name": "prediction", "role": "prediction", "type": "numerical"},
        ]
    """
    if name not in _MONITOR_INSTANCES:
        if not BentoMLContainer.config.monitoring.enabled.get():
            monitor_klass = NoOpMonitor
        elif monitor_class is None or monitor_class == "default":
            logger.debug("No monitor class is provided, will use default monitor.")
            monitor_klass = DefaultMonitor
        elif monitor_class == "otlp":
            from .otlp import OTLPMonitor

            monitor_klass = OTLPMonitor
        elif isinstance(monitor_class, str):
            monitor_klass = LazyType["MonitorBase[t.Any]"](monitor_class).get_class()
        elif isinstance(monitor_class, type):
            monitor_klass = monitor_class
        else:
            logger.warning(
                "Invalid monitor class, will disable monitoring. Please check your configuration. Setting monitor class to NoOp."
            )
            monitor_klass = NoOpMonitor

        if monitor_options is None:
            monitor_options = {}

        _MONITOR_INSTANCES[name] = monitor_klass(name, **monitor_options)

    mon = _MONITOR_INSTANCES[name]
    mon.start_record()
    yield mon
    mon.stop_record()



from __future__ import annotations

import json
import socket
import typing as t

import click


@click.command()
@click.argument("bento_identifier", type=click.STRING, required=False, default=".")
@click.option(
    "--fd",
    type=click.INT,
    required=True,
    help="File descriptor of the socket to listen on",
)
@click.option(
    "--runner-map",
    type=click.STRING,
    envvar="BENTOML_RUNNER_MAP",
    help="JSON string of runners map, default sets to envars `BENTOML_RUNNER_MAP`",
)
@click.option(
    "--backlog", type=click.INT, default=2048, help="Backlog size for the socket"
)
@click.option(
    "--working-dir",
    type=click.Path(exists=True),
    help="Working directory for the API server",
)
@click.option(
    "--prometheus-dir",
    type=click.Path(exists=True),
    help="Required by prometheus to pass the metrics in multi-process mode",
)
@click.option(
    "--worker-id",
    required=False,
    type=click.INT,
    default=None,
    help="If set, start the server as a bare worker with the given worker ID. Otherwise start a standalone server with a supervisor process.",
)
@click.option(
    "--ssl-certfile",
    type=str,
    default=None,
    help="SSL certificate file",
)
@click.option(
    "--ssl-keyfile",
    type=str,
    default=None,
    help="SSL key file",
)
@click.option(
    "--ssl-keyfile-password",
    type=str,
    default=None,
    help="SSL keyfile password",
)
@click.option(
    "--ssl-version",
    type=int,
    default=None,
    help="SSL version to use (see stdlib 'ssl' module)",
)
@click.option(
    "--ssl-cert-reqs",
    type=int,
    default=None,
    help="Whether client certificate is required (see stdlib 'ssl' module)",
)
@click.option(
    "--ssl-ca-certs",
    type=str,
    default=None,
    help="CA certificates file",
)
@click.option(
    "--ssl-ciphers",
    type=str,
    default=None,
    help="Ciphers to use (see stdlib 'ssl' module)",
)
@click.option(
    "--development-mode",
    type=click.BOOL,
    help="Run the API server in development mode",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--timeout",
    type=click.INT,
    help="Specify the timeout for API server",
)
def main(
    bento_identifier: str,
    fd: int,
    runner_map: str | None,
    backlog: int,
    working_dir: str | None,
    worker_id: int,
    prometheus_dir: str | None,
    ssl_certfile: str | None,
    ssl_keyfile: str | None,
    ssl_keyfile_password: str | None,
    ssl_version: int | None,
    ssl_cert_reqs: int | None,
    ssl_ca_certs: str | None,
    ssl_ciphers: str | None,
    development_mode: bool,
    timeout: int | None,
):
    """
    Start a HTTP api server worker.
    """
    import psutil

    import bentoml
    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.context import component_context
    from bentoml._internal.log import configure_server_logging

    component_context.component_type = "api_server"
    component_context.component_index = worker_id
    configure_server_logging()

    if worker_id is None:
        # worker ID is not set; this server is running in standalone mode
        # and should not be concerned with the status of its runners
        BentoMLContainer.config.runner_probe.enabled.set(False)

    BentoMLContainer.development_mode.set(development_mode)
    if prometheus_dir is not None:
        BentoMLContainer.prometheus_multiproc_dir.set(prometheus_dir)

    if runner_map is not None:
        BentoMLContainer.remote_runner_mapping.set(json.loads(runner_map))
    if timeout is not None:
        BentoMLContainer.api_server_config.traffic.timeout.set(timeout)
    svc = bentoml.load(bento_identifier, working_dir=working_dir, standalone_load=True)

    # setup context
    component_context.component_name = svc.name
    if svc.tag is None:
        component_context.bento_name = svc.name
        component_context.bento_version = "not available"
    else:
        component_context.bento_name = svc.tag.name
        component_context.bento_version = svc.tag.version or "not available"

    uvicorn_options: dict[str, t.Any] = {
        "backlog": backlog,
        "log_config": None,
        "workers": 1,
    }

    if ssl_certfile:
        import ssl

        uvicorn_options["ssl_certfile"] = ssl_certfile
        if ssl_keyfile:
            uvicorn_options["ssl_keyfile"] = ssl_keyfile
        if ssl_keyfile_password:
            uvicorn_options["ssl_keyfile_password"] = ssl_keyfile_password
        if ssl_ca_certs:
            uvicorn_options["ssl_ca_certs"] = ssl_ca_certs

        if not ssl_version:
            ssl_version = ssl.PROTOCOL_TLS_SERVER
            uvicorn_options["ssl_version"] = ssl_version
        if not ssl_cert_reqs:
            ssl_cert_reqs = ssl.CERT_NONE
            uvicorn_options["ssl_cert_reqs"] = ssl_cert_reqs
        if not ssl_ciphers:
            ssl_ciphers = "TLSv1"
            uvicorn_options["ssl_ciphers"] = ssl_ciphers

    if psutil.WINDOWS:
        uvicorn_options["loop"] = "asyncio"
        import asyncio

        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore

    import uvicorn

    # skip the uvicorn internal supervisor
    sock = socket.socket(fileno=fd)
    config = uvicorn.Config(svc.asgi_app, **uvicorn_options)
    uvicorn.Server(config).run(sockets=[sock])


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
BentoML
=======

BentoML is the unified ML Model Serving framework. Data Scientists and ML Engineers use
BentoML to:

* Accelerate and standardize the process of taking ML models to production across teams
* Build reliable, scalable, and high performance model serving systems
* Provide a flexible MLOps platform that grows with your Data Science needs

To learn more, visit BentoML documentation at: http://docs.bentoml.com
To get involved with the development, find us on GitHub: https://github.com/bentoml
And join us in the BentoML slack community: https://l.bentoml.com/join-slack
"""

from typing import TYPE_CHECKING

from ._internal.configuration import BENTOML_VERSION as __version__
from ._internal.configuration import load_config
from ._internal.configuration import save_config
from ._internal.configuration import set_serialization_strategy

# Inject dependencies and configurations
load_config()

# BentoML built-in types
from ._internal.bento import Bento
from ._internal.cloud import YataiClient
from ._internal.context import ServiceContext as Context
from ._internal.models import Model
from ._internal.monitoring import monitor
from ._internal.resource import Resource
from ._internal.runner import Runnable
from ._internal.runner import Runner
from ._internal.runner.strategy import Strategy
from ._internal.service import Service
from ._internal.service.loader import load
from ._internal.tag import Tag
from ._internal.utils.http import Cookie

# Bento management APIs
from .bentos import delete
from .bentos import export_bento
from .bentos import get
from .bentos import import_bento
from .bentos import list  # pylint: disable=W0622
from .bentos import pull
from .bentos import push
from .bentos import serve

# server API
from .server import GrpcServer
from .server import HTTPServer

# Framework specific modules, model management and IO APIs are lazily loaded upon import.
if TYPE_CHECKING:
    from . import catboost
    from . import detectron
    from . import diffusers
    from . import easyocr
    from . import fastai
    from . import flax
    from . import gluon
    from . import h2o
    from . import keras
    from . import lightgbm
    from . import mlflow
    from . import onnx
    from . import onnxmlir
    from . import paddle
    from . import picklable_model
    from . import pycaret
    from . import pytorch
    from . import pytorch_lightning
    from . import ray
    from . import sklearn
    from . import spacy
    from . import statsmodels
    from . import tensorflow
    from . import tensorflow_v1
    from . import torchscript
    from . import transformers
    from . import triton
    from . import xgboost

    # isort: off
    from . import io
    from . import models
    from . import metrics  # Prometheus metrics client
    from . import container  # Container API
    from . import client  # Client API
    from . import batch  # Batch API
    from . import exceptions  # BentoML exceptions
    from . import server  # Server API
    from . import monitoring  # Monitoring API
    from . import cloud  # Cloud API

    # isort: on
else:
    from ._internal.utils import LazyLoader as _LazyLoader

    # ML Frameworks
    catboost = _LazyLoader("bentoml.catboost", globals(), "bentoml.catboost")
    detectron = _LazyLoader("bentoml.detectron", globals(), "bentoml.detectron")
    diffusers = _LazyLoader("bentoml.diffusers", globals(), "bentoml.diffusers")
    easyocr = _LazyLoader("bentoml.easyocr", globals(), "bentoml.easyocr")
    flax = _LazyLoader("bentoml.flax", globals(), "bentoml.flax")
    fastai = _LazyLoader("bentoml.fastai", globals(), "bentoml.fastai")
    gluon = _LazyLoader("bentoml.gluon", globals(), "bentoml.gluon")
    h2o = _LazyLoader("bentoml.h2o", globals(), "bentoml.h2o")
    lightgbm = _LazyLoader("bentoml.lightgbm", globals(), "bentoml.lightgbm")
    mlflow = _LazyLoader("bentoml.mlflow", globals(), "bentoml.mlflow")
    onnx = _LazyLoader("bentoml.onnx", globals(), "bentoml.onnx")
    onnxmlir = _LazyLoader("bentoml.onnxmlir", globals(), "bentoml.onnxmlir")
    keras = _LazyLoader("bentoml.keras", globals(), "bentoml.keras")
    paddle = _LazyLoader("bentoml.paddle", globals(), "bentoml.paddle")
    pycaret = _LazyLoader("bentoml.pycaret", globals(), "bentoml.pycaret")
    pytorch = _LazyLoader("bentoml.pytorch", globals(), "bentoml.pytorch")
    pytorch_lightning = _LazyLoader(
        "bentoml.pytorch_lightning", globals(), "bentoml.pytorch_lightning"
    )
    sklearn = _LazyLoader("bentoml.sklearn", globals(), "bentoml.sklearn")
    picklable_model = _LazyLoader(
        "bentoml.picklable_model", globals(), "bentoml.picklable_model"
    )
    spacy = _LazyLoader("bentoml.spacy", globals(), "bentoml.spacy")
    statsmodels = _LazyLoader("bentoml.statsmodels", globals(), "bentoml.statsmodels")
    tensorflow = _LazyLoader("bentoml.tensorflow", globals(), "bentoml.tensorflow")
    tensorflow_v1 = _LazyLoader(
        "bentoml.tensorflow_v1", globals(), "bentoml.tensorflow_v1"
    )
    torchscript = _LazyLoader("bentoml.torchscript", globals(), "bentoml.torchscript")
    transformers = _LazyLoader(
        "bentoml.transformers", globals(), "bentoml.transformers"
    )
    xgboost = _LazyLoader("bentoml.xgboost", globals(), "bentoml.xgboost")

    # Integrations
    triton = _LazyLoader("bentoml.triton", globals(), "bentoml.triton")
    ray = _LazyLoader("bentoml.ray", globals(), "bentoml.ray")

    io = _LazyLoader("bentoml.io", globals(), "bentoml.io")
    batch = _LazyLoader("bentoml.batch", globals(), "bentoml.batch")
    models = _LazyLoader("bentoml.models", globals(), "bentoml.models")
    metrics = _LazyLoader("bentoml.metrics", globals(), "bentoml.metrics")
    container = _LazyLoader("bentoml.container", globals(), "bentoml.container")
    client = _LazyLoader("bentoml.client", globals(), "bentoml.client")
    server = _LazyLoader("bentoml.server", globals(), "bentoml.server")
    exceptions = _LazyLoader("bentoml.exceptions", globals(), "bentoml.exceptions")
    monitoring = _LazyLoader("bentoml.monitoring", globals(), "bentoml.monitoring")
    cloud = _LazyLoader("bentoml.cloud", globals(), "bentoml.cloud")

    del _LazyLoader

__all__ = [
    "__version__",
    "Context",
    "Cookie",
    "Service",
    "models",
    "batch",
    "metrics",
    "container",
    "client",
    "server",
    "io",
    "Tag",
    "Model",
    "Runner",
    "Runnable",
    "monitoring",
    "YataiClient",  # Yatai REST API Client
    # bento APIs
    "list",
    "get",
    "delete",
    "import_bento",
    "export_bento",
    "load",
    "push",
    "pull",
    "serve",
    "Bento",
    "exceptions",
    # server APIs
    "HTTPServer",
    "GrpcServer",
    # Framework specific modules
    "catboost",
    "detectron",
    "diffusers",
    "easyocr",
    "flax",
    "fastai",
    "gluon",
    "h2o",
    "lightgbm",
    "mlflow",
    "onnx",
    "onnxmlir",
    "paddle",
    "picklable_model",
    "pycaret",
    "pytorch",
    "pytorch_lightning",
    "keras",
    "sklearn",
    "spacy",
    "statsmodels",
    "tensorflow",
    "tensorflow_v1",
    "torchscript",
    "transformers",
    "xgboost",
    # integrations
    "ray",
    "cloud",
    "triton",
    "monitor",
    "load_config",
    "save_config",
    "set_serialization_strategy",
    "Strategy",
    "Resource",
]



from __future__ import annotations

import importlib.metadata
import importlib.util
from importlib.metadata import PackageNotFoundError
from types import ModuleType
from typing import cast

from packaging.version import Version

__all__ = [
    "PackageNotFoundError",
    "pkg_version_info",
    "get_pkg_version",
    "source_locations",
    "find_spec",
]

get_pkg_version = importlib.metadata.version
find_spec = importlib.util.find_spec


def pkg_version_info(pkg_name: str | ModuleType) -> tuple[int, int, int]:
    if isinstance(pkg_name, ModuleType):
        pkg_name = pkg_name.__name__
    pkg_version = Version(get_pkg_version(pkg_name))
    return pkg_version.major, pkg_version.minor, pkg_version.micro


def source_locations(pkg: str) -> str | None:
    module = find_spec(pkg)
    if module is None:
        return
    (module_path,) = module.submodule_search_locations  # type: ignore (unfinished typed)
    return cast(str, module_path)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import types

LATEST_PROTOCOL_VERSION = "v1"


def import_generated_stubs(
    version: str = LATEST_PROTOCOL_VERSION,
    file: str = "service.proto",
) -> tuple[types.ModuleType, types.ModuleType]:
    """
    Import generated stubs.

    Args:
        version: The version of the proto file to import.
        file: The name of the proto file to import.

    Returns:
        A tuple of the generated stubs for the proto file.

    Examples:

        .. code-block:: python

           from bentoml.grpc.utils import import_generated_stubs

           # given proto file bentoml/grpc/v1/service.proto exists
           pb, services = import_generated_stubs(version="v1", file="service.proto")
    """
    # generate git root from this file's path
    from bentoml._internal.utils import LazyLoader

    exception_message = f"Generated stubs for '{version}/{file}' are missing (broken installation). Please reinstall bentoml: 'pip install bentoml[grpc].'"
    file = file.split(".")[0]

    service_pb2 = LazyLoader(
        f"{file}_pb2",
        globals(),
        f"bentoml.grpc.{version}.{file}_pb2",
        exc_msg=exception_message,
    )
    service_pb2_grpc = LazyLoader(
        f"{file}_pb2_grpc",
        globals(),
        f"bentoml.grpc.{version}.{file}_pb2_grpc",
        exc_msg=exception_message,
    )
    return service_pb2, service_pb2_grpc


def import_grpc() -> tuple[types.ModuleType, types.ModuleType]:
    from bentoml._internal.utils import LazyLoader

    exception_message = "'grpcio' is required for gRPC support. Install with 'pip install bentoml[grpc]'."
    grpc = LazyLoader("grpc", globals(), "grpc", exc_msg=exception_message)
    aio = LazyLoader("aio", globals(), "grpc.aio", exc_msg=exception_message)
    return grpc, aio



from __future__ import annotations

INF = float("inf")

DEFAULT_BUCKET = (
    0.005,
    0.01,
    0.025,
    0.05,
    0.075,
    0.1,
    0.25,
    0.5,
    0.75,
    1.0,
    2.5,
    5.0,
    7.5,
    10.0,
    INF,
)

MAX_BUCKET_COUNT = 100


def metric_name(*args: str | int) -> str:
    """
    Concatenates the given parts into a legal Prometheus metric name. For example,
    a valid tag name may includes invalid characters, so we need to escape them
    ref: https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels
    """
    return "_".join([str(arg).replace("-", ":").replace(".", "::") for arg in args])


def exponential_buckets(start: float, factor: float, end: float) -> tuple[float, ...]:
    """
    Creates buckets of a Prometheus histogram where the lowest bucket has an upper
    bound of start and the upper bound of each following bucket is factor times the
    previous buckets upper bound. The return tuple include the end as the second
    last value and positive infinity as the last value.
    """

    assert start > 0.0
    assert start < end
    assert factor > 1.0

    bound = start
    buckets: list[float] = []
    while bound < end:
        buckets.append(bound)
        bound *= factor

    if len(buckets) > MAX_BUCKET_COUNT:
        buckets = buckets[:MAX_BUCKET_COUNT]

    return tuple(buckets) + (end, INF)


def linear_buckets(start: float, step: float, end: float) -> tuple[float, ...]:
    """
    Creates buckets of a Prometheus histogram where the lowest bucket has an upper
    bound of start and the upper bound of each following bucket is the previous
    buckets upper bound plus step. The return tuple include the end as the second
    last value and positive infinity as the last value.
    """

    assert start > 0.0
    assert start < end
    assert step > 0.0

    bound = start
    buckets: list[float] = []
    while bound < end:
        buckets.append(bound)
        bound += step

    if len(buckets) > MAX_BUCKET_COUNT:
        buckets = buckets[:MAX_BUCKET_COUNT]

    return tuple(buckets) + (end, INF)



from __future__ import annotations

import logging
import os
from pathlib import Path

from ...exceptions import BentoMLException
from ...exceptions import MissingDependencyException
from ...grpc.utils import LATEST_PROTOCOL_VERSION
from ..configuration import is_pypi_installed_bentoml
from ..utils.pkg import source_locations

logger = logging.getLogger(__name__)

BENTOML_DEV_BUILD = "BENTOML_BUNDLE_LOCAL_BUILD"


def build_bentoml_editable_wheel(
    target_path: str, *, _internal_protocol_version: str = LATEST_PROTOCOL_VERSION
) -> None:
    """
    This is for BentoML developers to create Bentos that contains the local bentoml
    build based on their development branch. To enable this behavior, one must
    set envar :code:`BENTOML_BUNDLE_LOCAL_BUILD=True` before building a Bento.
    """
    if str(os.environ.get(BENTOML_DEV_BUILD, False)).lower() != "true":
        return

    if is_pypi_installed_bentoml():
        # skip this entirely if BentoML is installed from PyPI
        return

    try:
        from build import ProjectBuilder
        from build.env import IsolatedEnvBuilder
    except ModuleNotFoundError as e:
        raise MissingDependencyException(
            f"Environment variable '{BENTOML_DEV_BUILD}=True', which requires the 'pypa/build' package ({e}). Install it with 'pip install -U build' and try again."
        ) from None

    # Find bentoml module path
    # This will be $GIT_ROOT/src/bentoml
    module_location = source_locations("bentoml")
    if not module_location:
        raise BentoMLException("Could not find bentoml module location.")
    bentoml_path = Path(module_location)

    if not Path(
        module_location, "grpc", _internal_protocol_version, "service_pb2.py"
    ).exists():
        raise ModuleNotFoundError(
            f"Generated stubs for version {_internal_protocol_version} are missing. Make sure to run '{bentoml_path.as_posix()}/scripts/generate_grpc_stubs.sh {_internal_protocol_version}' beforehand to generate gRPC stubs."
        ) from None

    # location to pyproject.toml
    pyproject = bentoml_path.parent.parent / "pyproject.toml"

    # this is for BentoML developer to create Service containing custom development
    # branches of BentoML library, it is True only when BentoML module is installed
    # in development mode via "pip install --editable ."
    if os.path.isfile(pyproject):
        logger.debug(
            "BentoML is installed in `editable` mode; building BentoML distribution with the local BentoML code base. The built wheel file will be included in the target bento."
        )
        with IsolatedEnvBuilder() as env:
            builder = ProjectBuilder(pyproject.parent)
            builder.python_executable = env.executable
            builder.scripts_dir = env.scripts_dir
            env.install(builder.build_system_requires)
            builder.build(
                "wheel", target_path, config_settings={"--global-option": "--quiet"}
            )
    else:
        logger.warning(
            "Custom BentoML build is detected. For a Bento to use the same build at serving time, add your custom BentoML build to the pip packages list, e.g. `packages=['git+https://github.com/bentoml/bentoml.git@13dfb36']`"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    import starlette.responses


@attr.define
class Cookie:
    key: str
    value: str
    max_age: int | None
    expires: int | None
    path: str
    domain: str | None
    secure: bool
    httponly: bool
    samesite: str


def set_cookies(response: starlette.responses.Response, cookies: list[Cookie]):
    for cookie in cookies:
        response.set_cookie(
            cookie.key,
            cookie.value,
            max_age=cookie.max_age,  # type: ignore (bad starlette types)
            expires=cookie.expires,  # type: ignore (bad starlette types)
            path=cookie.path,
            domain=cookie.domain,  # type: ignore (bad starlette types)
            secure=cookie.secure,
            httponly=cookie.httponly,
            samesite=cookie.samesite,
        )



from __future__ import annotations

import importlib
import logging
import typing as t

from ...exceptions import MissingDependencyException
from .api import monitor
from .base import MonitorBase
from .base import NoOpMonitor
from .default import DefaultMonitor

logger = logging.getLogger(__name__)

_is_otlp_available = False
try:
    from .otlp import OTLPMonitor as OTLPMonitor

    _is_otlp_available = True
except (ImportError, MissingDependencyException):
    # NOTE: we want this here so that this class will be imported as a dummy object.
    pass

__all__ = [
    "monitor",
    "MonitorBase",
    "DefaultMonitor",
    "NoOpMonitor",
]

if _is_otlp_available:
    __all__.append("OTLPMonitor")


def __getattr__(item: str) -> t.Any:
    if item == "OTLPMonitor" and not _is_otlp_available:
        raise MissingDependencyException(
            "OTLPMonitor is not available, please install it with `pip install bentoml[monitor-otlp]`"
        )
    elif item in __all__:
        return importlib.import_module(f".{item}", __name__)
    else:
        raise AttributeError(f"module {__name__} has no attribute {item}")
